// Lane.cpp: implementation of the Lane class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Lane.h"

#include "ConfigData.h"
extern CConfigData g_ConfigData;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

	/**Default constructor*/
Lane::Lane()
{
	m_LeftLane = NULL;
	m_RightLane = NULL;
}

	/**Default destructor*/
Lane::~Lane()
{

}

	/**Main constructor
	 *
	 * The main constructor used for creating a Lane object.
	 *
	 * @param iLane The lane's index
	 * @param RoadLength The length of the Road
	 * @param detectors The detectors to be used in the lane
	 * @param DirPos Whether the lane is in the positive direction or not
	 * @param AllowLaneChange Whether lane changing is allowed
	 */
Lane::Lane(int iLane, int RoadLength, std::vector<Detector*> detectors, bool DirPos, bool AllowLaneChange)
{
	
	MIN_SPACE_FOR_NEXT_VEHICLE = g_ConfigData.IDM.MIN_SPACE_FOR_NEXT_VEHICLE;
	m_PrevVehiclePosition = MIN_SPACE_FOR_NEXT_VEHICLE;
	
	m_NoVeh = 0;
	m_RoadLength = RoadLength;
	m_LeftLane = NULL;
	m_RightLane = NULL;
	m_bTrackLaneChanges = false;

	m_iLane = iLane;
	m_vDetectors = detectors;
	m_DirPos = DirPos;
	m_AllowLaneChanging = AllowLaneChange;

	for(int i = 0; i < m_vDetectors.size(); i++)
	{
		Detector* pDet = m_vDetectors.at(i);
		if( (pDet->getDetectorType() == METRICS_TYPE_LANE_CHANGE) && (pDet->getDirPos() == m_DirPos) )
		{
			m_bTrackLaneChanges = true;
			m_pLaneChangeDetector = dynamic_cast<LaneChangeDetector*>(pDet);
			break;
		}
	}
}

	/**Updates the positions of all Vehicles in the Lane
	 *
	 * This function is called for each iteration of the simulation, and handles the
	 * updating of the positions of Vehicles, and the removal of Vehicles which are
	 * no longer on the Road.
	 * If there are any Vehicles on the Road, the Vehicle at the head of the Lane is
	 * updated with respect to the clear Road ahead of it. Following this, all Vehicles 
	 * following the Vehicle at the head of the Road are updated with respect to the 
	 * preceding Vehicle. These Vehicles are also checked to see if it
	 * is time to determine if changing to another Lane would be advantageous.
	 * After all Vehicles have been updated, it is checked to see whether the Vehicle
	 * at the head of the Lane has yet passed the end of the Road. If it has, it is
	 * removed from the Lane and the simulation.
	 * Vehicles are also checked to see if they have passed a detector during the update
	 *
	 * @param step The timestep
	 * @param curTime The current simulation time
	 * @return Whether the lane is empty or not
	 * @see ChangeLanes()
	 * @see CheckFeatures()
	 */

// returns laneEmpty = true if no vehicles in lane
bool Lane::update(const double step, const double curTime)
{
	m_CurTime = curTime;
	bool laneEmpty = true;

	if(m_pVehicles.size() > 0)
	{
		// Update the head of the lane to drive towards the end of the road
		double previousPos = m_pVehicles.front()->getPos();
		m_pVehicles.front()->update(step);
		double currentPos = m_pVehicles.front()->getPos();
		CheckFeatures(curTime, previousPos, currentPos, m_pVehicles.front());

		// Update each vehicle based on the properties of the preceding vehicle
		for(int i = 1; i < m_pVehicles.size(); i++)
		{
			previousPos = m_pVehicles[i]->getPos();
			m_pVehicles[i]->update(step, m_pVehicles[i-1]);
			currentPos = m_pVehicles[i]->getPos();
			
			bool bChangedLane = false;
			if(m_AllowLaneChanging)											// Check if overtaking is allowed
			{
				bool bChangeStatus = m_pVehicles.at(i)->getChangeStatus();	// Ready to change lane?
				if(bChangeStatus)
				{
					Lane* pNewLane = ChangeLanes(m_pVehicles.at(i));		// return new lane if it's changed
					bChangedLane = pNewLane == NULL ? false : true;			// Has it changed?
					if(bChangedLane)
					{
						// get the new lane to check the vehicle for a detector/segment for this time step
						pNewLane->CheckFeatures(curTime, previousPos, currentPos, m_pVehicles.at(i));
						m_pVehicles.at(i)->setChangeStatus(false);			// Reset status
						m_pVehicles.erase(m_pVehicles.begin()+i);			// Remove from this lane
					}																			
				}
			}
			if(!bChangedLane)
				CheckFeatures(curTime, previousPos, currentPos, m_pVehicles.at(i));

		}
		// update position of last vehicle in lane
		// so more vehicles can get on the road
		m_PrevVehiclePosition = m_pVehicles.back()->getPos();	

		//Remove any vehicles that are no longer on the road
		if(m_pVehicles.at(0)->getPos() > m_RoadLength + MIN_SPACE_FOR_NEXT_VEHICLE)
		{
			delete m_pVehicles.front();
			m_pVehicles.erase(m_pVehicles.begin());
			m_NoVeh--;
		}
		laneEmpty = false;;
	}
	return laneEmpty;
}

// this is just a wrapper function and is public so a lane can tell
// another lane to check when a vehicle changes lane

	/** Checks all features on the road
	 *
	 * @param curTime The current time
	 * @param prevPosition The vehicle's position in the previous step
	 * @param curPosition The vehicle's current position
	 * @param pVeh The vehicle
	 * @see CheckDetectors()
	 * @see CheckRoadSegments()
	 *
	 * This function checks all of the special features on the road
	 */
void Lane::CheckFeatures(const double curTime, double prevPosition, double curPosition, Vehicle* pVeh)
{
	CheckDetectors(curTime, prevPosition, curPosition, pVeh);
	CheckRoadSegments(prevPosition, curPosition, pVeh);
}

	/** Adds a vehicle to the detectors list of vehicles for that step
	 *
	 * @param curTime The current simulation time
	 * @param prevPosition The vehicle's previous position
	 * @param curPosition The vehicle's current position
	 * @param pVeh The vehicle
	 *
	 * This function takes account of all detectors that are currently on the road
	 * and checks to see if a given vehicle has passed any detectors in the last
	 * simulation step. 
	 * If a vehicle has passed a detector, it is added to that detector
	 */
void Lane::CheckDetectors(const double curTime, double prevPosition, double curPosition, Vehicle* pVeh)
{
	// if the vehicle has moved past it in the last time step
	// add the vehicle to the detector
	
	//////// OUTPUT DETECTOR //////////////
	int DetectorLocation = m_pOutputDetector->getLocation();
	if(prevPosition <= DetectorLocation && curPosition >= DetectorLocation)
		m_pOutputDetector->addVehicle(pVeh, curTime);

	//////// METRICS DETECTORS ////////////
	
	for(int i = 0; i < m_vDetectors.size(); i++)
	{
		DetectorLocation = m_vDetectors.at(i)->getLocation();
		if(prevPosition <= DetectorLocation && curPosition >= DetectorLocation)
			m_vDetectors.at(i)->addVehicle(pVeh, curTime);
	}
}

	/**Gets the length of the Lane
	 *
	 * @return The length of the Lane
	 */
	//Probably not needed?

double Lane::getLength()
{
	return m_RoadLength;
}

	/**Gets all the Vehicles currently in the Lane
	 *
	 * This function returns the vector which contains pointers to all
	 * Vehicles which are currently in the Lane
	 *
	 * @return All of the Vehicles in the Lane
	 */
std::vector<Vehicle*> Lane::getPos()
{
	return m_pVehicles;
}

	/**Sets a pointer to the Lane to the left of the current Lane
	 * The Lane index is in global coords:
	 * - For DriveOnRight the lanes are numbered from bottom to top 0...n
	 * - For left driving the lanes are numbered from top to bottom 0...n
	 *
	 * @param left The Lane to point to
	 */
void Lane::setLeftLane(Lane *left)
{
	m_LeftLane = left;
}

	/**Sets a pointer to the Lane to the right of the current Lane
	 * The Lane index is in global coords:
	 * - For DriveOnRight the lanes are numbered from bottom to top 0...n
	 * - For left driving the lanes are numbered from top to bottom 0...n
	 *
	 * @param right The Lane to point to
	 */
void Lane::setRightLane(Lane *right)
{
	m_RightLane = right;
}

	/** Gets the number of vehicles in the lane
	 *
	 * @return The number of vehicles in the lane
	 */
int Lane::getNoVeh()
{
	return m_pVehicles.size();
}

	/** Gets the position of the last vehicle in the lane
	 *
	 * @return The position of the last vehicle in the lane
	 */
double Lane::getLastPos()
{
	return m_PrevVehiclePosition;
}
	
	/** Gets the index of the lane
	 *
	 * @return The index of the lane
	 */
int Lane::getIndex()
{
	return m_iLane;
}

	/**Changes a Vehicle's Lane if advantageous and possible
	 *
	 * @param pVeh The vehicle changing lanes
	 * @return Whether or not the vehicle has changed lane
	 *
	 * This function is called whenever it is time for a Vehicle to check whether
	 * or not it would be advantageous to move to another Lane. The vehicle finds
	 * its neighbours in the left and right lanes, and then decides whether to
	 * change lane, or which lane to change to, based on their properties.
	 *
	 */

Lane* Lane::ChangeLanes(Vehicle *pVeh)
{
	// since if it's a single lane you cannot change so let's not even bother
	if(m_RightLane == NULL && m_LeftLane == NULL)
		return false;

	bool DirPos = pVeh->getDirection();
	Vehicle* LeftFrontVehicle = NULL;	Vehicle* LeftBackVehicle = NULL;
	Vehicle* RightFrontVehicle = NULL;  Vehicle* RightBackVehicle = NULL;
	int iLeftLane = 0;					int iRightLane = 0;
	bool LeftLaneExists = m_LeftLane != NULL ? true : false;
	bool RightLaneExists = m_RightLane != NULL ? true : false;
	
	if(LeftLaneExists)
	{
		LeftFrontVehicle = FinadAdjacentVehicle(m_LeftLane, pVeh->getPos(), true, &iLeftLane);
		LeftBackVehicle = FinadAdjacentVehicle(m_LeftLane, pVeh->getPos(), false);
	}
	
	if(RightLaneExists)
	{
		RightFrontVehicle = FinadAdjacentVehicle(m_RightLane, pVeh->getPos(), true, &iRightLane);
		RightBackVehicle = FinadAdjacentVehicle(m_RightLane, pVeh->getPos(), false);
	}
	int LaneChangeID = 0;
	if(DirPos)	// in DirPos
		LaneChangeID = pVeh->DecideLaneChange(	LeftFrontVehicle,	LeftBackVehicle, 
												RightFrontVehicle,	RightBackVehicle,
												LeftLaneExists,		RightLaneExists);
	else	// in DirNeg so swap vehicles - see comment in Vehicle::DecideLaneChange()
		LaneChangeID = pVeh->DecideLaneChange(	RightFrontVehicle,	RightBackVehicle,
												LeftFrontVehicle,	LeftBackVehicle,
												RightLaneExists,	LeftLaneExists);

	if(LaneChangeID == 0)
		return NULL;
	else
	{
		Lane* pNewLane = ImplementLaneChange(pVeh, LaneChangeID, iLeftLane, iRightLane);
		return pNewLane;
	}
}

	/** Changes a vehicle to another lane
	 *
	 * @param pVeh The vehicle changing lane
	 * @param LaneChangeID The lane to change to
	 * @param iVehLeftLane The index of the vehicle if moving to the left lane
	 * @param iVehRightLane The index of the vehicle if moving to the right lane
	 *
	 * This function inserts the vehicle that is changing lane into the appropriate lane. It also
	 * takes care of tracking lane changes if the user specifies that lane changes should be
	 * tracked.
	 */
Lane* Lane::ImplementLaneChange(Vehicle* pVeh, int LaneChangeID, int iVehLeftLane, int iVehRightLane)
{
	Lane* VehiclesLeftLane;		Lane* VehiclesRightLane;	Lane* LaneToChange;
	
	int fromLane = pVeh->getLane();		// for lane change detector
	double position = pVeh->getPos();
	bool DirPos = pVeh->getDirection();
	bool changeLeft = true;

	if(DirPos)
	{
		VehiclesLeftLane = m_LeftLane;		VehiclesRightLane = m_RightLane;
	}
	else
	{
		VehiclesLeftLane = m_RightLane;		VehiclesRightLane = m_LeftLane;

		int temp = iVehLeftLane;			// Swapping vehicle index in new Lane
		iVehLeftLane = iVehRightLane;
		iVehRightLane = temp;

		position = m_RoadLength - position;
	}
	if(LaneChangeID == 1)
	{
		// change to the vehicle's left lane
		pVeh->setLane(VehiclesLeftLane->getIndex()+1);	// vehicle lane numbers are 1-based
		VehiclesLeftLane->insert(iVehLeftLane, pVeh);
		LaneToChange = VehiclesLeftLane;
	}
	else
	{
		changeLeft = false;
		// change to vehicle's right lane
		pVeh->setLane(VehiclesRightLane->getIndex()+1);	// vehicle lane numbers are 1-based
		VehiclesRightLane->insert(iVehRightLane, pVeh);
		LaneToChange = VehiclesRightLane;
	}

	if(m_bTrackLaneChanges)	// If we are tracking changes, add this event
	{
		int toLane = (LaneToChange->getIndex() + 1);
		LaneChangeEvent* lce = new LaneChangeEvent(position, fromLane, toLane, pVeh->getID(), changeLeft, DirPos, m_CurTime);
		m_pLaneChangeDetector->addEvent(lce);
	}

	return LaneToChange;
}

	/** Finds a vehicle in the adjacent lane
	 *
	 * @param AdjacentLane The lane at which we are looking
	 * @param location The location of the vehicle wishing to change
	 * @param front Whether or not we are looking for the vehicle in front of the current vehicle's position
	 * @param iAdjacentLane The index to switch into in the adjacent lane
	 * @return The vehicle desired
	 *
	 * This function finds the vehicle that in the adjacent lane, in front of (front == true) the given
	 * location, or behind (front == false) the given location. It returns the appropriate vehicle, and 
	 * sets iAdjacentIndex to the index into which the current vehicle should change if it changes lane.
	 */
Vehicle* Lane::FinadAdjacentVehicle(Lane* AdjacentLane, double location, bool front, int* iAdjacentLane)
{
	std::vector<Vehicle*> AdjacentLaneVeh = AdjacentLane->getPos();

	if(AdjacentLaneVeh.size() == 0)
		return NULL;

	// searches from back of nextLane up to the vehicle in front
	int i = AdjacentLaneVeh.size() - 1;

	while(i >= 0 && AdjacentLaneVeh.at(i)->getPos() < location)
		i--;

	if(iAdjacentLane != NULL)	// i points to the vehicle in front of us in the other lane
		*iAdjacentLane = i+1;	// place the position in iAdjacentLane

	//We need to check for bounds here, since we're returning a Vehicle, not an index

	if(front && i >= 0)
		return AdjacentLaneVeh.at(i);
	else if(front && i < 0)
		return NULL;
	else if(!front && i == AdjacentLaneVeh.size() - 1)
		return NULL;
	else
		return AdjacentLaneVeh.at(i+1);
}

	/**Inserts a Vehicle into the Lane
	 *
	 * This function inserts a given Vehicle into a particular position in
	 * a Lane
	 *
	 * @param i The position to insert the Vehicle into
	 * @param pVeh The Vehicle to insert
	 */
void Lane::insert(int i, Vehicle *pVeh)
{
	if(i == m_pVehicles.size())	// insert at end
	{
		m_pVehicles.push_back(pVeh);
		m_PrevVehiclePosition = pVeh->getPos();
	}
	else if(i <= 0)				// insert at front
		m_pVehicles.insert(m_pVehicles.begin(), pVeh);
	else						// insert between front & back, i.e. lane change
		m_pVehicles.insert(m_pVehicles.begin()+i, pVeh);

	m_NoVeh++;
}

	/** Sets the output detector for the direction
	 *
	 * @param out The output detector
	 */
void Lane::setOutputDetector(OutputDetector *out)
{
	m_pOutputDetector = out;
}

	/** Sets the road segments for the direction
	 *
	 * @param segments The road segments
	 */
void Lane::setRoadSegments(std::vector<RoadSegment*> segments)
{
	m_RoadSegments = segments;
}

	/**Checks whether a vehicle has entered or exited any road segments
	 *
	 * @param prevPosition The position of the vehicle before the update
	 * @param curPosition The position of the vehicle after the update
	 * @param pVeh The vehicle involved
	 *
	 * This function takes account of all segments in the road - speedlimits, gradients, etc - 
	 * and checks each update to see if a vehicle has entered or exited any of the segments on
	 * the road. 
	 * Overlapping segments are accounted for, as if a vehicle exits a segment and 
	 * has not entered a subsequent segment, it is checked to see whether the vehicle is still
	 * in the bounds of a previous segment
	 */
void Lane::CheckRoadSegments(double prevPosition, double curPosition, Vehicle *pVeh)
{
	bool entered = false;
	bool exited = false;
	int nSegments = m_RoadSegments.size();

	for(int i = 0; i < nSegments; i++)						// for each road segment
	{
		int start = m_RoadSegments.at(i)->getBeginning();	// its start and end
		int end = m_RoadSegments.at(i)->getEnd();

		if(prevPosition <= start && curPosition >= start)	// if we have entered the segment
		{
			m_RoadSegments.at(i)->addVehicle(pVeh);
			entered = true;
		}
		else if(prevPosition < end && curPosition >= end)	// else if we have exited the segment
		{
			m_RoadSegments.at(i)->removeVehicle(pVeh);
			exited = true;
		}
	}

	if(exited && !entered) // if we have exited a segment, but not entered a subsequent one
	{
		for(int j = 0; j < nSegments; j++)
		{
			int start = m_RoadSegments.at(j)->getBeginning();
			int end = m_RoadSegments.at(j)->getEnd();

			if(curPosition >= start && curPosition < end)
			{
				m_RoadSegments.at(j)->addVehicle(pVeh);
				entered = true;
				break;
			}
		}
		if(!entered) 
			pVeh->getDriver()->ClearRestriction();
	}
}

	/** Clears the lane so it can be used in another simulation
	 *
	 * This function clears all of the lane's information so that it may
	 * be used in another simulation without re-instanciating.
	 */
void Lane::clear()
{
	
	// If we exit in mid-sim, clear all vehicles currently on the road
	for(int i = 0; i < m_pVehicles.size(); i++)
		delete m_pVehicles.at(i);

	// Free the vector for vehicles
	m_pVehicles.clear();
	
	// Free the vectors for detectors and segments (both deleted in direction)
	m_vDetectors.clear();
	m_RoadSegments.clear();
}
