// MetricsDetector.cpp: implementation of the MetricsDetector class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MetricsDetector.h"


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

	/**Constructor
	 *
	 * @param dir The detector output directory
	 * @param DetType The particular type of detector
	 * @param DirPos Whether the detector is in the positive direction
	 * @param nLanes The number of lanes the detector covers
	 * @param VehType The type of vehicles to track
	 * @param timeInterval The time interval to use
	 * @param loc The location of the detector
	 * @param RoadLength The length of the road
	 */
MetricsDetector::MetricsDetector(std::string dir, WORD DetType, bool DirPos, int nLanes,
								 WORD VehType, int timeInterval, int loc, int RoadLength)
{
	m_MetricsDir = dir;
	m_DetectorType = DetType;
	m_DirPos = DirPos;
	m_NoLanes = nLanes;
	m_VehicleType = VehType;
	m_TimeInterval = timeInterval;
	m_Location = loc;
	m_RoadLength = RoadLength;

	m_CurrentTime = 0;
	m_IntervalNo = 0;
	
	m_vTimeOfLastArrival.assign(m_NoLanes,0.0);	// vector of last arrival times by lane
	initVectors();
	InitOutputFiles();
}

	/**Destructor*/
MetricsDetector::~MetricsDetector()
{

}
	/**Do the output at the end of each time interval*/
void MetricsDetector::doIntervalOutput()
{
	switch(m_DetectorType)
	{
		case METRICS_TYPE_FLOWDENSITY:
			doFlowDensityOutput();
			break;
		case METRICS_TYPE_HEADWAY:
			doHeadwayOutput();
			break;
		case METRICS_TYPE_COMPOSITION:
			if(m_VehicleType == METRICS_VEH_ALL)
				doCompositionOutput();	// since composition only makes sense with all vehicles
			break;
		default:
			doFlowDensityOutput();
	}

	ReInit();
}
	/**Check if we are to add the current vehicle to the detector*/
void MetricsDetector::addVehicle(Vehicle* pVeh, double curTime)
{
	WORD vehID = pVeh->getID();

	switch(m_VehicleType)
	{
		case METRICS_VEH_ALL:			
			AddCurrentVehicle(pVeh, curTime);
			break;
		case METRICS_VEH_CAR:
			if(vehID == VEH_ID_CAR)
				AddCurrentVehicle(pVeh, curTime);
			break;
		case METRICS_VEH_SMALLTRUCK:
			if(vehID == VEH_ID_SMALLTRUCK)
				AddCurrentVehicle(pVeh, curTime);
			break;
		case METRICS_VEH_LARGETRUCK:
			if(vehID == VEH_ID_LARGETRUCK)
				AddCurrentVehicle(pVeh, curTime);
			break;
		case METRICS_VEH_CRANE:
			if(vehID == VEH_ID_CRANE)
				AddCurrentVehicle(pVeh, curTime);
			break;
		case METRICS_VEH_LOWLOADER:
			if(vehID == VEH_ID_LOWLOADER)
				AddCurrentVehicle(pVeh, curTime);
			break;
		default:
			AddCurrentVehicle(pVeh, curTime);
	}

}

// in this function we add the vehicle's properties as appropriate
// depending on the mode this detector is working in

	/**Add the current vehicle to the detector*/
void MetricsDetector::AddCurrentVehicle(Vehicle* pVeh, const double curTime)
{
	int iLane = pVeh->getLaneNoInDirection();	// get the local lane number in its direction
	switch(m_DetectorType)
	{
		case METRICS_TYPE_FLOWDENSITY:
		{
			m_vNoVehicles[iLane-1]++;									// increase vehicle count by lane
			m_vVelocities[iLane-1].push_back(pVeh->getVelocity());		// and store velocity
			break;
		}
		case METRICS_TYPE_HEADWAY:
		{
			double ArrivalTime = FindDetectorArrivalTime(pVeh, curTime);
			double headway = ArrivalTime - m_vTimeOfLastArrival[iLane-1];
			m_vHeadways[iLane-1].push_back(headway);
			m_vTimeOfLastArrival[iLane-1] = ArrivalTime;	// update last arrival time for this lane
			break;
		}
		case METRICS_TYPE_COMPOSITION:
		{
			// since composition only makes sense with all vehicles
			if(m_VehicleType == METRICS_VEH_ALL)
				m_vVehicleTypes.push_back( pVeh->getID() );
			break;
		}
	}
	
}

	/**At the end of each time interval reset vectors for next time period*/
void MetricsDetector::ReInit()
{
	m_vVehicleTypes.clear();
	m_vNoVehicles.clear();
	for(int i = 0; i < m_NoLanes; i++)
	{
		m_vVelocities[i].clear();
		m_vHeadways[i].clear();
	}
	m_vVelocities.clear();
	m_vHeadways.clear();
		
	initVectors();
}

	/**Initialise storage vectors*/
void MetricsDetector::initVectors()
{
	m_vNoVehicles.assign(m_NoLanes,0);			// no of vehicles by lane
	std::vector<double> temp;
	m_vHeadways.assign(m_NoLanes, temp);
	m_vVelocities.assign(m_NoLanes, temp);
}

	/**Initialise the output files*/
void MetricsDetector::InitOutputFiles()
{
	switch(m_DetectorType)
	{
		case METRICS_TYPE_FLOWDENSITY:
			InitFlowDensityFile();
			break;
		case METRICS_TYPE_HEADWAY:
			InitHeadwayFile();
			break;
		case METRICS_TYPE_COMPOSITION:
			if(m_VehicleType == METRICS_VEH_ALL)	// since composition only makes sense with all vehicles
				InitCompositionFile();
			break;
		default:
			InitFlowDensityFile();
	}
}

	/**Initialise the Flow-Density output file*/
void MetricsDetector::InitFlowDensityFile()
{	
	// Create file name and open it
	int loc = m_DirPos == true ? m_Location : m_RoadLength - m_Location;
	std::string file;
	file = m_MetricsDir;
	file += "FD_" + MapDetVehTypeToString(m_VehicleType) + "_" 
				 + MapDirToString(m_DirPos) +"_" 
				 + intToString(loc) + ".csv";
	m_OutFile.open(file.c_str(), std::ios::out);
	
	// let's set some headers
	m_OutFile << " " << ',';	//Interval no column
	for(int i = 1; i <= m_NoLanes; i++)
		m_OutFile << "Lane " <<  intToString(i) << ", , ,";
	m_OutFile << "Totals" << '\n';

	m_OutFile << "Interval No."		<< ',';
	for(i = 0; i < m_NoLanes+1; i++)	// + 1 is for the totals columns
	{
		char end = i == m_NoLanes ? '\n' : ',';
		m_OutFile << "Density (veh/km)"	<< ',';
		m_OutFile << "Flow (veh/hr)"	<< ',';
		m_OutFile << "Avg. Vel. (km/h)"	<< end;
	}
}

	/**Initialise the Headway output file*/
void MetricsDetector::InitHeadwayFile()
{	
	// Create file name and open it
	int loc = m_DirPos == true ? m_Location : m_RoadLength - m_Location;
	std::string file;
	file = m_MetricsDir;
	file += "Head_" + MapDetVehTypeToString(m_VehicleType) + "_" 
				   + MapDirToString(m_DirPos) +"_" 
				   + intToString(loc) + ".csv";
	m_OutFile.open(file.c_str(), std::ios::out);
	
	// let's set some headers
	m_OutFile << "Headways (s)"	<< '\n';
	for(int i = 1; i < m_NoLanes; i++)
		m_OutFile << "Lane " <<  intToString(i) << ',';
	m_OutFile << "Lane " <<  intToString(m_NoLanes) << '\n';
}

	/**Initialise the Composition output file*/
void MetricsDetector::InitCompositionFile()
{	
	// Create file name and open it
	int loc = m_DirPos == true ? m_Location : m_RoadLength - m_Location;
	std::string file;
	file = m_MetricsDir;
	file += "Comp_" + MapDetVehTypeToString(m_VehicleType) + "_" 
				   + MapDirToString(m_DirPos) +"_" 
				   + intToString(loc) + ".csv";
	m_OutFile.open(file.c_str(), std::ios::out);
	
	// let's set some headers
	m_OutFile << "No. Cars"			<< ",";
	m_OutFile << "No. Small Trucks"	<< ","; 
	m_OutFile << "No. Large Trucks"	<< ","; 
	m_OutFile << "No. Cranes"		<< ","; 
	m_OutFile << "No. Low-Loaders"	<< '\n'; 
}

	/**Do the Flow-Density output*/
void MetricsDetector::doFlowDensityOutput()
{
	// first find the density, flow and ave velocity
	std::vector<double> vAvgVel(m_NoLanes,0.0);
	std::vector<double> vFlowPerHour(m_NoLanes,0.0);
	std::vector<double> vDensity(m_NoLanes,0.0);

	for(int i = 0; i < m_NoLanes; i++)
	{
		for(int j = 0; j < m_vNoVehicles[i]; j++)
			vAvgVel[i] += m_vVelocities[i].at(j);
		// since we have the sum of velocites in lane i now 
		// we can get the average now, in km/h
		vAvgVel[i] = M_PER_S_TO_KM_PER_H * vAvgVel[i] / m_vNoVehicles[i];
		vFlowPerHour[i] = m_vNoVehicles[i] * (SECS_PER_HOUR/m_TimeInterval);
		vDensity[i] = vFlowPerHour[i]/vAvgVel[i];	// units of veh/km
	}
	
	// lastly, do the sum of lanes for total direction properties
	double totAvgVel = 0.0; double totFlowPerHour = 0.0; double totDensity = 0.0;
	for(i = 0; i < m_NoLanes; i++)
	{
		totAvgVel		+= vAvgVel[i];
		totFlowPerHour	+= vFlowPerHour[i];
		totDensity		+= vDensity[i];
	}
		
	// then do the output and reset the variables
	m_OutFile << m_IntervalNo	<< ',';
	for(i = 0; i < m_NoLanes; i++)
	{
		m_OutFile << vDensity[i]		<< ',';
		m_OutFile << vFlowPerHour[i]	<< ',';
		m_OutFile << vAvgVel[i]			<< ',';
	}
	m_OutFile << totDensity		<< ',';
	m_OutFile << totFlowPerHour	<< ',';
	m_OutFile << totAvgVel		<< '\n';
}

	/**Do the Headway output*/
void MetricsDetector::doHeadwayOutput()
{
	// get size of largest headway values vector
	int jmax = m_vHeadways[0].size();
	for(int i = 1; i < m_NoLanes; i++)
	{
		if(m_vHeadways[i].size() > jmax)
			jmax = m_vHeadways[i].size();
	}

	// write across then down, leaving blanks where no value
	for(int j = 0; j < jmax; j++)
	{
		for(i = 0; i < m_NoLanes; i++)
		{
			char end = i == m_NoLanes-1 ? '\n' : ',';

			if(j < m_vHeadways[i].size())
				m_OutFile << m_vHeadways[i].at(j) << end;
			else
				m_OutFile << " " << end;
		}
	}
}

	/**Do the Composition output*/
void MetricsDetector::doCompositionOutput()
{
	int nCar = 0;
	int nSmallTruck = 0;
	int nLargeTruck = 0;
	int nCrane = 0;
	int nLowLoader = 0;

	for(int i = 0; i < m_vVehicleTypes.size(); i++)
	{
		WORD VehType = m_vVehicleTypes.at(i);

		switch(VehType)
		{
			case VEH_ID_CAR:		nCar++;			break;
			case VEH_ID_SMALLTRUCK:	nSmallTruck++;	break;
			case VEH_ID_LARGETRUCK:	nLargeTruck++;	break;
			case VEH_ID_CRANE:		nCrane++;		break;
			case VEH_ID_LOWLOADER:	nLowLoader++;	break;
			default: nCar++; 
		}
	}	

	m_OutFile << nCar			<< ",";
	m_OutFile << nSmallTruck	<< ","; 
	m_OutFile << nLargeTruck	<< ","; 
	m_OutFile << nCrane			<< ","; 
	m_OutFile << nLowLoader		<< '\n'; 
}

	/**Change an int to a string object*/
std::string MetricsDetector::intToString(int d)
{
	std::string a;
	char buffer; char* pBuffer = &buffer;
	itoa( d, pBuffer, 10 );
	a.append(&buffer);
	return a;
}

	/**Close the output file*/
void MetricsDetector::EndOutput()
{
	m_OutFile.close();
}
