
#include "stdafx.h"
#include "Road.h"

#include "ConfigData.h"
extern CConfigData g_ConfigData;

	/** Constructor */
Road::Road()
{
	MIN_SPACE_FOR_NEXT_VEHICLE = g_ConfigData.IDM.MIN_SPACE_FOR_NEXT_VEHICLE;

	m_pIDMParams_Car		= NULL;
	m_pIDMParams_SmallTruck = NULL;
	m_pIDMParams_LargeTruck = NULL;
	m_pIDMParams_Crane		= NULL;
	m_pIDMParams_Lowloader	= NULL;

	m_pFileHandler			= NULL;

	m_bEndOfFile = false;
}

	/** Destructor */
Road::~Road()
{
	m_pIDMParams_Car		= NULL;
	m_pIDMParams_SmallTruck = NULL;
	m_pIDMParams_LargeTruck = NULL;
	m_pIDMParams_Crane		= NULL;
	m_pIDMParams_Lowloader	= NULL;
}

	/** Updates the properties of the road
	 *
	 * @param step The timestep
	 * @param curTime The current time
	 * @return Whether or not the simulation is still running
	 *
	 * This function is called for each timestep of the simulation and updates its
	 * directions, once the directions are not empty
	 */
bool Road::update(double step,const double curTime)
{
	bool SimulationOver = false;
	bool RoadEmpty = false;
	bool DirPosEmpty = false;
	bool DirNegEmpty = false;

	// since there's always one direction
	DirPosEmpty = DirectionPos.update(step, curTime);

	if(m_NoDirections == 2)
		DirNegEmpty = DirectionNeg.update(step, curTime);

	if(DirPosEmpty && DirNegEmpty)
		RoadEmpty = true;

	if(RoadEmpty && m_bEndOfFile)	// if no more vehicles and road is empty
		SimulationOver = true;		// we are finished!

	if(SimulationOver)	// tidy up detector output
	{
		for(int i = 0; i < m_vDetectorsPos.size(); i++)
			m_vDetectorsPos[i]->EndOutput();
		for(i = 0; i < m_vDetectorsNeg.size(); i++)
			m_vDetectorsNeg[i]->EndOutput();
	}

	return SimulationOver;			
}

	/** Initialises the road
	 *
	 * @return The time to the first vehicle
	 *
	 * This function initialises the road object. Firstly, the special segments of the road
	 * are sorted depending on their road position. Next, we load a buffer of vehicles which 
	 * the road processes, and maps these vehicles to their respective lanes. 
	 * Directions are initialised, and the road then takes account of cumulative lane identifiers 
	 * by loading each direction's lanes into a vector of cumulative lanes.
	 * Detectors, if specified by the user, are also set in each direction to facilitate the 
	 * retrieval of updated vehicle information when a vehicle passes a specified point. 
	 */
double Road::init()
{	
	sort(m_vRoadSegmentsPos.begin(), m_vRoadSegmentsPos.end(), RoadSegment::CompareSegments);	
	sort(m_vRoadSegmentsNeg.begin(), m_vRoadSegmentsNeg.end(), RoadSegment::CompareSegments);	

	// Initially set up vehicles and their sim lanes
	m_BufferVehicles = m_pFileHandler->loadTruckGroup(INPUT_FILE_BUFFER_SIZE);
	for(int i = 0; i < m_BufferVehicles.size(); i++)
		MapTrafLaneToSimLane(m_BufferVehicles.at(i));

	/////////////////// DIRECTION POSITIVE ////////////////////////////////
	// lane vector
	std::vector<int> vLaneLengthsPos(m_NoLanesDirPos,m_RoadLength);	// this is for future extension

	// true means positive x-direction
	OutputDetector* pDetDirPos = new OutputDetector(m_LocOutputDetectorDirPos, true, m_pFileHandler);
	DirectionPos.init(true, pDetDirPos, m_AllowLaneChanging, m_RoadLength, m_DriveOnRight);		
	DirectionPos.createLanes(0, vLaneLengthsPos, m_vDetectorsPos, m_vRoadSegmentsPos);
	
	// Using cumulative lane identifiers
	for(i = 0; i < DirectionPos.getNoLanes(); i++)
		m_vLanes.push_back(&DirectionPos.getLane(i));

	/////////////////// DIRECTION NEGATIVE ////////////////////////////////
	if(m_NoDirections == 2)
	{
		// lane vector
		std::vector<int> vLaneLengthsNeg(m_NoLanesDirNeg,m_RoadLength);
		
		// false means the negative x-direction
		OutputDetector* pDetDirNeg = new OutputDetector(m_LocOutputDetectorDirNeg, false, m_pFileHandler);
		DirectionNeg.init(false, pDetDirNeg, m_AllowLaneChanging, m_RoadLength, m_DriveOnRight);
		DirectionNeg.createLanes(m_NoLanesDirPos, vLaneLengthsNeg, m_vDetectorsNeg, m_vRoadSegmentsNeg);

		for(i = 0; i < DirectionNeg.getNoLanes(); i++)
			m_vLanes.push_back(&DirectionNeg.getLane(i));
	}

	// Return the time of the first truck of the traffic file
	return m_BufferVehicles.front()->getTime();
}

	/** Populates the road
	 *
	 * @param step The timestep
	 * @param curTime The simulation's current time
	 *
	 * This function handles the population of the road which is called in each
	 * timestep. The main loop handles the processing of the vehicles in the buffer.
	 * If the head of the buffer is currently entering the road, then the other 
	 * vehicles are also checked. However, since the vehicles are sorted in order of
	 * time, if the head of the buffer is not currently entering the road, then no
	 * other vehicle is either.
	 * When a vehicle enters the road from the buffer, the next vehicle is read from
	 * the file being used, and added to the buffer.
	 */
void Road::populate(double step, const double curTime)
{
	bool frontOn = true;
	while(m_BufferVehicles.size() > 0 && frontOn)
	{
		frontOn = false;
		int lane = m_BufferVehicles.front()->getLane() - 1;		// the cumulative lane no of the next truck on the road

		if(m_vLanes.at(lane)->getLastPos() >= MIN_SPACE_FOR_NEXT_VEHICLE)		// If there is space for another vehicle
		{
			if(m_BufferVehicles.front()->getTime() <= curTime)	// If it is time for the next vehicle
			{
				int noVeh = m_vLanes.at(lane)->getNoVeh();		// crashes here sim has fewer lanes than the truck file.

				// set the driver, roadlength, insert into lane and remove from buffer
				setIDMDriverModel(m_BufferVehicles.front());
				m_BufferVehicles.front()->init(m_RoadLength, m_DriveOnRight);
				m_vLanes.at(lane)->insert(noVeh, m_BufferVehicles.front());
				m_BufferVehicles.erase(m_BufferVehicles.begin());

				Vehicle* tempVeh = m_pFileHandler->readLine();
				m_PercentComplete = m_pFileHandler->getPercentComplete();
				if(tempVeh != NULL)
				{
					MapTrafLaneToSimLane(tempVeh);
					m_BufferVehicles.push_back(tempVeh);
				}
				else
					m_bEndOfFile = true;
				frontOn = true;
			}
		}

	}
}

	/** Maps vehicles to appropriate lanes
	 *
	 * @param pVeh The vehicle to map
	 *
	 * This function updates the lane provided by the vehicle's original traffic
	 * file assignment to an appropriate lane in the simulation. This is to
	 * facilitate the simulation of more lanes than the input file allows for
	 */
void Road::MapTrafLaneToSimLane(Vehicle* pVeh)
{
	// first tell the vehicle how many lanes the road has for later use
	pVeh->setTotalNoLanesInRoad(m_NoLanes);

	// then map its current local to global lane number
	if(!pVeh->getDirection())	// only if it's in the negative direction
	{
		int TrafFileNoLanes = m_TrafFileNoLanesDirPos + m_TrafFileNoLanesDirNeg;
		int lanesFromNegVerge = TrafFileNoLanes - pVeh->getLane();
		int newLane = m_NoLanes - lanesFromNegVerge;	// vehicle lane numbers are 1-based
		pVeh->setLane(newLane);
	}
}

	/** Sets a vehicle's drivermodel as IDM
	 *
	 * @param pVeh The vehicle to set
	 *
	 * This function sets the drivermodel of a particular vehicle based on the
	 * type of that vehicle
	 */
void Road::setIDMDriverModel(Vehicle* pVeh)
{
	WORD VEH_ID = pVeh->getID();
	
	switch(VEH_ID)
	{
		case VEH_ID_CAR:
			pVeh->setDriver( m_pIDMParams_Car->Generate() );
			break;
		case VEH_ID_SMALLTRUCK:
			pVeh->setDriver( m_pIDMParams_SmallTruck->Generate() );
			break;
		case VEH_ID_LARGETRUCK:
			pVeh->setDriver( m_pIDMParams_LargeTruck->Generate() );
			break;
		case VEH_ID_CRANE:
			pVeh->setDriver( m_pIDMParams_Crane->Generate() );
			break;
		case VEH_ID_LOWLOADER:
			pVeh->setDriver( m_pIDMParams_Lowloader->Generate() );
			break;
		default:
			pVeh->setDriver( m_pIDMParams_Car->Generate() );
	};
}

	/** Gets all the vehicles on the road
	 *
	 * @return All the vehicles on the road
	 *
	 */
M2D Road::getVehicles()
{	
	M2D pos;

	M2D temp = DirectionPos.getGlobalPos();	// get DirPos positions
	for(int i = 0; i < m_NoLanesDirPos; i++)		// and add lane by lane
		pos.push_back(temp[i]);

	if(m_NoDirections == 2)		// Uni-directional roads are in Dir1
	{
		temp.clear();
		
		temp = DirectionNeg.getGlobalPos();	// get DirNeg positions
		for(i = 0; i < m_NoLanesDirNeg; i++)		// and add lane by lane
			pos.push_back(temp[i]);
	}

	return pos;
}

	/** Gets the length of the road
	 *
	 * @return The length of the road
	 */
double Road::getLength()
{
	return m_RoadLength;
}

	/** Sets the number of lanes in the positive direction
	 *
	 * @param nlpos The number of lanes in the positive direction
	 */
void Road::setNoLanesDirPos(int nlpos)
{
	m_NoLanesDirPos = nlpos;
}

	/** Sets the number of lanes in the negative direction
	 *
	 * @param nlneg The number of lanes in the negative direction
	 */
void Road::setNoLanesDirNeg(int nlneg)
{
	m_NoLanesDirNeg = nlneg;
}

	/** Sets the number of directions
	 *
	 * @param nd The number of directions
	 */
void Road::setNoDirections(int nd)
{
	m_NoDirections = nd;
}

	/** Sets the number of lanes
	 *
	 * @param nl The number of lanes
	 */
void Road::setNoLanes(int nl)
{
	m_NoLanes = nl;
}

	/** Sets the length of the road
	 *
	 * @param L The length of the road
	 */
void Road::setRoadLength(int L)
{
	m_RoadLength = L;
}

	/** Initialises the file handler
	 *
	 * @param inFile The input file to read from
	 * @param outFile The output file to write to
	 * @param fileType The type of file
	 *
	 * This function determines whether the type of file being used is
	 * CASTOR or SAFT format, and creates an appropriate filehandler
	 * based on the input provided.
	 */
void Road::initTruckGroup(string inFile, string outFile, WORD fileType)
{
	if(fileType == CASTOR)
	{
		m_pFileHandler = new CASTORFile(inFile, outFile);

	}
	else if(fileType == SAFT)
	{
		m_pFileHandler = new SAFTFile(inFile, outFile);
	}
}

	/** Sets the Car IDM Parameter Set
	 *
	 * @param Params The IDM Parameter Set
	 */
void Road::setIDMParams_Car(CIDMParameterSet* Params)
{
	m_pIDMParams_Car = Params;
}

	/** Sets the Small Truck IDM Parameter Set
	 *
	 * @param Params The IDM Parameter Set
	 */
void Road::setIDMParams_SmallTruck(CIDMParameterSet* Params)
{
	m_pIDMParams_SmallTruck = Params;
}

	/** Sets the Large Truck IDM Parameter Set
	 *
	 * @param Params The IDM Parameter Set
	 */
void Road::setIDMParams_LargeTruck(CIDMParameterSet* Params)
{
	m_pIDMParams_LargeTruck = Params;
}

	/** Sets the Crane IDM Parameter Set
	 *
	 * @param Params The IDM Parameter Set
	 */
void Road::setIDMParams_Crane(CIDMParameterSet* Params)
{
	m_pIDMParams_Crane = Params;
}

	/** Sets the Lowloader IDM Parameter Set
	 *
	 * @param Params The IDM Parameter Set
	 */
void Road::setIDMParams_Lowloader(CIDMParameterSet* Params)
{
	m_pIDMParams_Lowloader = Params;
}

	/** Sets the number of lanes in the positive direction in the traffic file
	 *
	 * @param nl The number of lanes
	 */
void Road::setTrafFileNoLanesDirPos(int nl)
{
	m_TrafFileNoLanesDirPos = nl;
}

	/** Sets the number of lanes in the negative direction in the traffic file
	 *
	 * @param nl The number of lanes
	 */
void Road::setTrafFileNoLanesDirNeg(int nl)
{
	m_TrafFileNoLanesDirNeg = nl;
}

	/** Sets the position of output detector in the positive direction
	 *
	 * @param loc The position of output detector
	 */
void Road::setLocOuputDetectorDirPos(int loc)
{
	m_LocOutputDetectorDirPos = loc;
}

	/** Sets the position of output detector in the negative direction
	 *
	 * @param loc The position of output detector
	 *
	 */
void Road::setLocOuputDetectorDirNeg(int loc)
{
	// change from overall road coordinates to lane coordinates
	m_LocOutputDetectorDirNeg = m_RoadLength - loc;
}

int Road::getPercentComplete()
{
	return m_PercentComplete;
}

	/** Gets whether or not lane changing is allowed
	 *
	 * @return Whether or not lane changing is allowed
	 */
bool Road::getAllowLaneChanging()
{
	return m_AllowLaneChanging;
}

	/** Sets whether or not lane changing is allowed
	 *
	 * @param status Whether or not lane changing is allowed
	 */
void Road::setAllowLaneChanging(bool status)
{
	m_AllowLaneChanging = status;
}

	/** Sets whether or not the vehicles should drive on the right
	 *
	 * @param OnRight Whether or not the vehicles should drive on the right
	 */
void Road::setDriveOnRight(bool OnRight)
{
	m_DriveOnRight = OnRight;
}

	/** Passes the IDM parameters to a given road segment
	 *
	 * @param segment The segment to pass the parameters to
	 */
void Road::passIDMParameters(RoadSegment *segment)
{
	segment->setIDMParams_Car(m_pIDMParams_Car);
	segment->setIDMParams_Crane(m_pIDMParams_Crane);
	segment->setIDMParams_LargeTruck(m_pIDMParams_LargeTruck);
	segment->setIDMParams_Lowloader(m_pIDMParams_Lowloader);
	segment->setIDMParams_SmallTruck(m_pIDMParams_SmallTruck);
}

	/**Sets road segments based on data specified by the user to the GUI
	 *
	 * @param pFeat The road feature to create
	 *
	 * This function takes the data inputted by the user into the GUI and creates
	 * the correct road segment based on the given data
	 */
void Road::SetSegmentFromFeature(CRoadFeature *pFeat)
{
	ASSERT(pFeat->getStart() < pFeat->getEnd());	// should be since done in CRoadFeaturesDlg validation

	switch( pFeat->getType() )
	{
		case FEAT_SPEEDLIMIT:
			SetSpeedLimitSegment(pFeat);
			break;
		case FEAT_GRADIENT:	
			SetGradientSegment(pFeat);
			break;
		default:
			break;
	}
	
}

	/**Sets a speed limit based on data specified by the user to the GUI
	 *
	 * @param pFeat The speed limit to create
	 *
	 * This function takes the data inputted by the user into the GUI and creates
	 * a speed limit based on the given data
	 */
void Road::SetSpeedLimitSegment(CRoadFeature *pFeat)
{
	bool bDirPos = pFeat->getDirPos();
	SpeedLimit* pSL;
	if( bDirPos )
	{
		pSL = new SpeedLimit(pFeat->getStart(), pFeat->getEnd(), pFeat->getValue(), bDirPos);
		m_vRoadSegmentsPos.push_back(pSL);
	}
	else
	{
		int start = pFeat->getStart();
		int end = pFeat->getEnd();

		pSL = new SpeedLimit(m_RoadLength - end, m_RoadLength - start, pFeat->getValue(), bDirPos);
		
		m_vRoadSegmentsNeg.push_back(pSL);
	}
	passIDMParameters(pSL);
}

	/**Sets a gradient based on data specified by the user to the GUI
	 *
	 * @param pFeat The gradient to create
	 *
	 * This function takes the data inputted by the user into the GUI and creates
	 * a gradient based on the given data
	 */
void Road::SetGradientSegment(CRoadFeature *pFeat)
{
	bool bDirPos = pFeat->getDirPos();
	Gradient* pGr;

	if( bDirPos )
	{
		pGr = new Gradient(pFeat->getStart(), pFeat->getEnd(), pFeat->getValue(), bDirPos);
		m_vRoadSegmentsPos.push_back(pGr);
	}
	else
	{
		int start = pFeat->getStart();
		int end = pFeat->getEnd();

		pGr = new Gradient(m_RoadLength - end, m_RoadLength - start, pFeat->getValue(), bDirPos);
		
		m_vRoadSegmentsNeg.push_back(pGr);
	}
	passIDMParameters(pGr);
}

	/**Sets the road metrics detector data from input through the GUI
	 *
	 * @param pDet The detector to create
	 * @param nLanesPos The number of lanes in the positive direction
	 * @param nLanesNeg The number of lanes in the negative direction
	 *
	 * This function takes the data inputted by the user into the GUI and creates
	 * a detector based on the given data. The detector is then passed to the 
	 * correct direction.
	 */
void Road::SetMetricDetFromStatDet(CStatDetector* pDet, int nLanesPos, int nLanesNeg)
{
	bool bDirPos = pDet->getDirPos();

	if(pDet->getDetectorType() == METRICS_TYPE_LANE_CHANGE)
	{
		LaneChangeDetector* pLCD;
		pLCD = new LaneChangeDetector(	pDet->getStdStrMetricsDir(),
										pDet->getTimeInterval(),
										pDet->getLocation(),
										m_RoadLength,
										bDirPos,
										pDet->getVehicleType());
		if( bDirPos )
			m_vDetectorsPos.push_back(pLCD);
		else
			m_vDetectorsNeg.push_back(pLCD);
	}
	else
	{
		MetricsDetector* pMD;
		if( bDirPos )
		{
			pMD = new MetricsDetector(	pDet->getStdStrMetricsDir(),
										pDet->getDetectorType(), 
										bDirPos,
										nLanesPos,
										pDet->getVehicleType(), 
										pDet->getTimeInterval(), 
										pDet->getLocation(), 
										m_RoadLength);
			m_vDetectorsPos.push_back(pMD);
		}
		else
		{
			pMD = new MetricsDetector(	pDet->getStdStrMetricsDir(),
										pDet->getDetectorType(), 
										bDirPos,
										nLanesNeg,
										pDet->getVehicleType(), 
										pDet->getTimeInterval(), 
										m_RoadLength - pDet->getLocation(),	// change from global road to local dirn coords
										m_RoadLength);						// need this to revert filenames to global positions
			m_vDetectorsNeg.push_back(pMD);
		}
	}
}

	/** Clears the road so it can be used in another simulation
	 *
	 * This function clears all of the road's information so that it may
	 * be used in another simulation without re-instanciating.
	 */
void Road::clear()
{
	DirectionPos.clear();
	if(m_NoDirections == 2)
		DirectionNeg.clear();

	delete m_pFileHandler;

	m_vLanes.clear();

	m_BufferVehicles.clear();

	m_vDetectorsPos.clear();
	m_vDetectorsNeg.clear();

	m_vRoadSegmentsPos.clear();
	m_vRoadSegmentsNeg.clear();

	// why don't we just delete these?
	m_pIDMParams_Car		= NULL;
	m_pIDMParams_SmallTruck = NULL;
	m_pIDMParams_LargeTruck = NULL;
	m_pIDMParams_Crane		= NULL;
	m_pIDMParams_Lowloader	= NULL;
}
