// Vehicle.cpp: implementation of the Vehicle class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Vehicle.h"

#include "ConfigData.h"
extern CConfigData g_ConfigData;

	/** Default Constructor */
Vehicle::Vehicle()
{
	ROAD_END_BUFFER				= g_ConfigData.Road.ROAD_END_BUFFER;
	
	T_DELAY						= g_ConfigData.IDM.T_DELAY;
	MIN_SPACE_FOR_NEXT_VEHICLE	= g_ConfigData.IDM.MIN_SPACE_FOR_NEXT_VEHICLE;
	SAFE_BRAKING				= g_ConfigData.IDM.SAFE_BRAKING;

	CRANE_AVERAGE_SPACING		= g_ConfigData.VehicleID.CRANE_AVERAGE_SPACING;
	CRANE_MAX_SPACING			= g_ConfigData.VehicleID.CRANE_MAX_SPACING;
	LOWLOADER_MIN_MAX_SPACING	= g_ConfigData.VehicleID.LOWLOADER_MIN_MAX_SPACING;
	SMALL_TRUCK_NO_AXLES		= g_ConfigData.VehicleID.SMALL_TRUCK_NO_AXLES;

	DAYS_PER_MT					= g_ConfigData.Time.DAYS_PER_MT;
	MTS_PER_YR					= g_ConfigData.Time.MTS_PER_YR;
}

	/** Default Destructor */
Vehicle::~Vehicle()
{

}

//////////////////////////////////////////////////////////////////////////////////////
///////////////////////////// VEHICLE DRIVING FUNCTIONS //////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

	/** Updates the position of the Vehicle based on the timestep
	 * @param step the timestep
	 *
	 * The Vehicle calling this function will always be the head of traffic in
	 * its particular Lane, as it has no Vehicle in front to reference and instead
	 * drives towards the end of the Road
	 * This function also handles the timer for when the Vehicle will check to see if changing Lane
	 * would be advantageous.
	 */

void Vehicle::update(double step)
{
	if( changeTime(step) )
		setChangeStatus(true);

	double accel = calcAccel();

	updateProperties(step, accel);
}

	/** Updates the position of the Vehicle based on the timestep and preceding Vehicle
	 * @param step the timestep
	 * @param pV a pointer to the Vehicle in front of this one
	 *
	 * Every Vehicle in the Lane will call this function except the head of the Lane. The velocity
	 * and acceleration of the Vehicle will be altered based on the velocity of, and distance to,
	 * the Vehicle in front.
	 * This function also handles the timer for when the Vehicle will check to see if changing Lane
	 * would be advantageous.
	 */

void Vehicle::update(double step, Vehicle *pV)
{
	bool change = changeTime(step);

	if(change)
		setChangeStatus(true);

	double accel = calcAccel(pV);

	updateProperties(step, accel);
}

	/** Calculates a Vehicle's acceleration when unimpeded by other Vehicles
	 *
	 * This function serves the purpose of determining the acceleration of a Vehicle when
	 * there are no other Vehicles in front of it in the same Lane. Based on the distance
	 * to the end of the Road, the DriverModel alters acceleration accordingly
	 */
double Vehicle::calcAccel()
{
	m_pDriver->setVel(m_Velocity);

	//MAGIC NUMBER: ROAD END
	double accel = m_pDriver->update(m_Velocity, (m_RoadLength + ROAD_END_BUFFER) - m_Position); 

	return accel;
}

	/** Calculates a Vehicle's acceleration based on properties of the preceding Vehicle
	 * @param pV A pointer to the preceding Vehicle
	 *
	 * This function serves the purpose of determining the acceleration of a Vehicle when
	 * there is another Vehicle in front of it in the same Lane. Based on the distance
	 * between the two vehicles, and the difference in their velocity, the DriverModel
	 * alters acceleration accordingly.
	 */
double Vehicle::calcAccel(Vehicle *pV)
{
	m_pDriver->setVel(m_Velocity);

	double s = pV->getPos() - pV->getLength() - m_Position;

	double delta_v = m_Velocity - pV->getVelocity();
	double accel = m_pDriver->update(delta_v, s);

	return accel;
}

	/** Updates the velocity, acceleration and position of the Vehicle based its acceleration and the timestep
	 * @param step The timestep
	 * @param accel The Vehicle's acceleration
	 *
	 * This function updates the properties of the Vehicle based on how much it has accelerated in the previous
	 * timestep, assuming average velocity over the timestep. The velocity, and acceleration of the Vehicle are
	 * updated, as well as the Vehicle's position.
	 */

void Vehicle::updateProperties(double step, double accel)
{
	// Assume average velocity over timestep
	double v_old = m_Velocity;
	double v_new = m_Velocity + accel*step;
	
	// zero speeds may occur when very close to stationary obstacle
	// eliminate as per Treiber's comments to Carpool Tunnel	
	if(v_new < 0)
	{
		//TRACE("*** Backwards move prevented\n");
		v_new = 0.0;
	}

	m_Position += 0.5*(v_old + v_new)*step;

	m_RoadPosition = m_Position;
	m_Velocity = v_new;

	m_Acceleration = accel;
}

	/** Handles the timer for checking Lane changes
	 *
	 * @param step The timestep
	 * @return Whether or not it is time to check for a Lane change
	 */
bool Vehicle::changeTime(double step)
{
	m_Tdelay += step;

	if (m_Tdelay >= T_DELAY)
	{
		m_Tdelay = m_Tdelay - T_DELAY;
		return true;
	}
	else {return false;}
}

	/**Gets the Vehicle's DriverModel
	 * @return the Vehicle's driver
	 */
DriverModel* Vehicle::getDriver()
{
	 return m_pDriver;
}

	/**Sets the Vehicle's DriverModel
	 * @param driver The Vehicle's driver
	 */
void Vehicle::setDriver(IDM driver)
{
	m_IDMDriver = driver;
	m_pDriver = &m_IDMDriver;
}

//////////////////////////////////////////////////////////////////////////////////////
////////////////////////////// LANE CHANGING FUNCTIONS ///////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

	/** Handles whether or not the Vehicle will change Lane
	 *
	 * @param LeftFrontVehicle The Vehicle in front in the Left Lane
	 * @param LeftBackVehicle The Vehicle behind in the Left Lane
	 * @param RightFrontVehicle The Vehicle in front in the Right Lane
	 * @param RightBackVehicle The Vehicle behind in the Right Lane
	 * @param LeftLaneExists Whether or not there is a Lane to the left
	 * @param RightLaneExists Whether or not there is a Lane to the Right
	 * @return The Lane to change to
	 * @see LaneChangeAdvantage()
	 *
	 * This function determines which Lane the Vehicle should change to if it changes
	 * Lane at all. Pointers to Vehicles that would be in front of, and behind, the 
	 * current Vehicle if it were in either the right or left Lane are passed to the
	 * function, along with identifiers which specify whether or not there exists a 
	 * Lane to the left or right of the current Lane.
	 */
int Vehicle::DecideLaneChange(	Vehicle* LeftFrontVehicle,	Vehicle* LeftBackVehicle,
								Vehicle* RightFrontVehicle, Vehicle* RightBackVehicle,
								bool LeftLaneExists,		bool RightLaneExists)
{
	// Note that a possible conflict between what is the lane's leftLane and rightLane
	// and the Vehicle's leftLane and Rightlane. For DirPos they are the same, but not 
	// for DirNeg. e.g.
	// DirPos - Lane 1: leftLane = 2: rightLane = 0;
	// DirNeg - Lane 5: leftLane = 6: rightLane = 4
	// Whereas for the Vehicle:
	// DirPos - in lane 1: lane to left = 2: lane to right = 0;
	// DirNeg - in lane 5: lane to left = 4: lane to right = 6;
	// Therefore:
	// THIS FUNCTION NAMES LANES AS VIEWED BY THE VEHICLE
	
	double LeftLaneAdvantage = 0.0;		double RightLaneAdvantage = 0.0;

	if(m_DriveOnRight)
	{	// we are overtaking properly
		if(LeftLaneExists)
			LeftLaneAdvantage = LaneChangeAdvantage(LeftFrontVehicle, LeftBackVehicle, true);
		// or we are being bold
		if(RightLaneExists)
			RightLaneAdvantage = LaneChangeAdvantage(RightFrontVehicle, RightBackVehicle, false);
	}
	else
	{	// we're boldly undertaking
		if(LeftLaneExists)
			LeftLaneAdvantage = LaneChangeAdvantage(LeftFrontVehicle, LeftBackVehicle, false);
		// we are overtaking properly
		if(RightLaneExists)
			RightLaneAdvantage = LaneChangeAdvantage(RightFrontVehicle, RightBackVehicle, true);
	}

	const double zero = 0.0001;
	if(LeftLaneAdvantage < zero && RightLaneAdvantage < zero)
		return 0;									// do not change lane
	else if(LeftLaneAdvantage > RightLaneAdvantage)
		return 1;									// move to the left lane
	else
		return 2;									// move to the right lane
}

	/**Gets the advantage a Vehicle would gain from switching Lanes
	 *
	 * @param FrontVehicle The Vehicle in front in the new Lane
	 * @param BackVehicle The Vehicle behind in the new Lane
	 * @param overtake Whether the Vehicle is overtaking or undertaking
	 * @return The net advantage of the Lane change
	 *
	 * This function takes pointers to the Vehicles both in front, and behind, the position 
	 * that the current Vehicle would have in the new Lane. If the Vehicle that would be in 
	 * front is null, then the change in acceleration is decided with relation to acceleration
	 * on an open Road. Otherwise, the acceleration is decided with relation to the Vehicle in
	 * front.
	 */
double Vehicle::LaneChangeAdvantage(Vehicle* FrontVehicle, Vehicle* BackVehicle, bool overtake)
{
	// These are IDM properties and may be reomved from Constants.h
	double GapToFront = MIN_SPACE_FOR_NEXT_VEHICLE;			double GapToBack = MIN_SPACE_FOR_NEXT_VEHICLE; 
	double CurrentBackAccel	= SAFE_BRAKING;					double ProposedBackAccel= SAFE_BRAKING;
	double FrontChangeAccel	= 0.0;

	// calculate the above lane change parameters
	if(FrontVehicle != NULL)
	{
		GapToFront = FrontVehicle->getPos() - FrontVehicle->getLength() - this->getPos();
		FrontChangeAccel = this->calcAccel(FrontVehicle) - this->getAccel();
	}
	else
		FrontChangeAccel = this->calcAccel() - this->getAccel();	// no front Vehicle

	if(BackVehicle != NULL)
	{
		GapToBack = this->getPos() - this->getLength() - BackVehicle->getPos();
		ProposedBackAccel = BackVehicle->calcAccel(this);
		// Since the acceleration of the back Vehicle at the next step
		// is affected by the presence of the Vehicle in front of it
		// we must take that into account and not an open-road acceleration
		// but only if there is a Vehicle in front!
		if(FrontVehicle != NULL)
			CurrentBackAccel = BackVehicle->calcAccel(FrontVehicle);
		else
			CurrentBackAccel = BackVehicle->getAccel();
	}

	double advantage =  m_pDriver->LaneChange(GapToFront, GapToBack, FrontChangeAccel, 
												CurrentBackAccel, ProposedBackAccel, overtake);
	return advantage;
}


//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////	VEHICLE PROPERTIES GETS & SETS ///////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

	/** Sets the velocity of the Vehicle to a given value
	 * @param vel the Vehicle's new velocity
	 */

void Vehicle::setVelocity(double vel)
{
	m_Velocity = vel;
}

	/** Sets the length of the Vehicle to a given value
	 * @param length The Vehicle's new length
	 */

void Vehicle::setLength(double length)
{
	m_Length = length;
}

	/** Sets the acceleration of the Vehicle to a given value
	 * @param acc the Vehicle's new acceleration
	 */

void Vehicle::setAccel(double acc)
{
	m_Acceleration = acc;
}

	/** Sets the lane that the Vehicle is in
	 * @param lane the Vehicle's lane
	 */

void Vehicle::setLane(int lane)
{
	m_Lane = lane;
}

	/** Sets the direction of the Vehicle
	 *
	 * @param DirPos Whether or not the vehicle is travelling in a positive direction
	 */

void Vehicle::setDirection(bool DirPos)
{
	m_DirPos = DirPos;
}


	/** Sets the position of the Vehicle to a given value
	 * @param p the Vehicle's position
	 */

void Vehicle::setPos(double p)
{
	m_Position = p;
}

	/** Sets whether or not the Vehicle wishes to change lanes
	 * @param stat whether or not the Vehicle wants to change lanes
	 */

void Vehicle::setChangeStatus(bool stat)
{
	m_bLaneChange = stat;
}

	/** Sets the position of the Vehicle on the road. For graphical use
	 * @param r the Vehicle's position on the road
	 */

void Vehicle::setRoadPos(double r)
{
	m_RoadPosition = r;
}


	/** Gets the Vehicle's position
	 * @return m_x the Vehicle's position
	 */

double Vehicle::getPos()
{
	return m_Position;
}

	/** Gets the desired velocity of the Vehicle's driver model
	 * @return the driver model's desired velocity
	 */

double Vehicle::getDesiredVel()
{
	return m_pDriver->getDesiredVel();
}

	/** Gets the position of the Vehicle on the road. Graphical use
	 * @return roadPos the position of the Vehicle on the road
	 */

double Vehicle::getRoadPos()
{
	return m_RoadPosition;
}

	/** Gets the class of the Vehicle
	 * @return ID the Vehicle's class
	 */

WORD Vehicle::getID()
{
	return m_ID;
}

	/** Gets whether or not the Vehicle wishes to change lane
	 * @return laneChange whether or not the Vehicle wishes to change lane
	 */

bool Vehicle::getChangeStatus()
{
	return m_bLaneChange;
}

	/** Gets the velocity of the Vehicle
	 * @return vel the velocity of the Vehicle
	 */

double Vehicle::getVelocity()
{
	return m_Velocity;
}

	/** Gets the length of the Vehicle
	 * @return len the length of the Vehicle
	 */

double Vehicle::getLength()
{
	return m_Length;
}

	/** Gets the acceleration of the Vehicle
	 * @return acc the acceleration of the Vehicle
	 */

double Vehicle::getAccel()
{
	return m_Acceleration;
}

	/** Gets the lane that the Vehicle is in
	 * @return lane the Vehicle's lane
	 */

int Vehicle::getLane()
{
	return m_Lane;
}

	/** Gets the direction that the Vehicle is travelling in
	 * @return dir the Vehicle's direction
	 */

bool Vehicle::getDirection()
{
	return m_DirPos;
}

	/** Gets the GVW of the vehicle
	 * @return the GVW of the vehicle
	 */

double Vehicle::getGVW()
{
	return m_GVW;
}

	/**Sets the GVW of the vehicle to a given weight
	 * @param weight the desired GVW
	 */
void Vehicle::setGVW(double weight)
{
	m_GVW = weight;
}

WORD Vehicle::setID()
{
	// Small Truck or large truck
	m_ID = (m_NoAxles <= SMALL_TRUCK_NO_AXLES) ? 
						VEH_ID_SMALLTRUCK : VEH_ID_LARGETRUCK;


	double averageSpace = getAS(0);
	double maxSpace = 0;

	// For all the vehicle's axles
	for(int i = 1; i < m_NoAxles; i++)
	{
		// Obtain the average spacing
		averageSpace += getAS(i);

		// Obtain the maximum spacing
		if(getAS(i) > getAS(i-1))
			maxSpace = getAS(i);
	}

	// If the maximum spacing is less than the maximum spacing for a crane and there are more than 2 axles
	if(maxSpace < CRANE_MAX_SPACING && m_NoAxles > 2)
	{
		if(averageSpace / m_vAxles.size() < CRANE_AVERAGE_SPACING) // And the average spacing is less than the maximum average for a crane
			m_ID = VEH_ID_CRANE;	// It is a crane
	}

	// If the maximum spacing is at least the maximum for a low-loader and there are more than 2 axles
	if(maxSpace >= LOWLOADER_MIN_MAX_SPACING && m_NoAxles > 2)
		m_ID = VEH_ID_LOWLOADER;	//It is a low-loader

	return m_ID;
}

	/** Gets the time of arrival of the vehicle in seconds
	 *
	 * @return The time of arrival of the vehicle in seconds
	 */

double Vehicle::getTime() const
{
	// working off a 10 month year & a 25 day month
	int s_per_hr = SECS_PER_HOUR;
	int s_per_day = SECS_PER_HOUR*HOURS_PER_DAY;

	int no_days = (m_Year - 5) * MTS_PER_YR * DAYS_PER_MT + (m_Month - 1) * DAYS_PER_MT + (m_Day-1);
	double time = no_days * s_per_day + m_Hour * s_per_hr + m_Min * SECS_PER_MIN + m_Sec + (double)m_Hndt/100;

	return time;
}

void Vehicle::setTime(double time)
{
	double temp = time;	
	const int mpy = MTS_PER_YR;	const int dpm = DAYS_PER_MT;

	m_Year	= (int)(temp/(MTS_PER_YR * DAYS_PER_MT * HOURS_PER_DAY * SECS_PER_HOUR));
	
	temp	= time - m_Year * MTS_PER_YR * DAYS_PER_MT * HOURS_PER_DAY * SECS_PER_HOUR;	
	m_Month	= (int)(temp/(DAYS_PER_MT * HOURS_PER_DAY * SECS_PER_HOUR));
	
	temp	= temp - m_Month * DAYS_PER_MT * HOURS_PER_DAY * SECS_PER_HOUR;	
	m_Day	= (int)(temp/(HOURS_PER_DAY * SECS_PER_HOUR));		
	
	temp	= temp - m_Day * HOURS_PER_DAY * SECS_PER_HOUR;			
	m_Hour	= (int)(temp/(SECS_PER_HOUR));
	
	temp	= temp - m_Hour * SECS_PER_HOUR;					
	m_Min	= (int)(temp/(MINS_PER_HOUR));
	
	temp	= temp - m_Min * MINS_PER_HOUR;						
	m_Sec	= (int)(temp);
	
	temp	= temp - m_Sec;							
	m_Hndt	= (int)(100 * temp);

}

void Vehicle::init(int RoadLength, bool DriveOnRight)
{
	m_RoadLength = RoadLength;
	m_DriveOnRight = DriveOnRight;
}

void Vehicle::setRoadLength(int length)
{
	m_RoadLength = length;
}

	/** Gets the vehicle's number of axles
	 * @return the vehicle's number of axles
	 */

int Vehicle::getNoAxles()
{
	return m_NoAxles;
}

	/** Sets the spacing of a given axle
	 *
	 * The space of a given axle to the next axle
	 * @param i the axle to operate upon
	 * @param s the desired spacing
	 */

void Vehicle::setAS(int i, double s)
{
	m_vAxles.at(i)->setSpacing(s);
}

	/** Sets the weight of a given axle
	 * @param i the axle to operate upon
	 * @param w the desired weight
	 */
void Vehicle::setAW(int i, double w)
{
	m_vAxles.at(i)->setWeight(w);
}

	/**Sets the vehicle's number of axles to a given number
	 *
	 * This function sets the vehicle's number of Axles and current differentiates
	 * between large and small trucks based on the number of Axles
	 */
void Vehicle::setAxles()
{
	for(int i = 0; i < m_NoAxles; i++)
	{
		Axle* axle = new Axle();
		m_vAxles.push_back(axle);
	}
}

	/** Gets the spacing of a given axle
	 *
	 * The space of a given axle to the next axle
	 * @param i the axle to operate upon
	 * @return the space between this axle and the next
	 */

double Vehicle::getAS(int i)
{
	return m_vAxles.at(i)->getSpacing();
}

	/** Gets the weight of a given axle
	 * @param i the axle to operate upon
	 * @return the weight of the axle
	 */

double Vehicle::getAW(int i)
{
	return m_vAxles.at(i)->getWeight();
}


void Vehicle::setTotalNoLanesInRoad(int nLanes)
{
	m_TotalNoLanesInRoad = nLanes;
}

// map global lane to local lane
int Vehicle::getLaneNoInDirection()
{
	if(m_DirPos)		// if pos direction, local = global
		return m_Lane;
	else				// if neg, it's not
	{
		ASSERT(m_Lane <= m_TotalNoLanesInRoad); // if it isn't we have a negative index returned
		int iLocalLane = m_TotalNoLanesInRoad - m_Lane + 1;	// must be 1-based lane index for file output
		return iLocalLane;
	}
}

//////////////////////////////////////////////////////////////////////////////////////
///////////////////////	FILE READ & WRITE VEHICLE FUNCTIONS //////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

void Vehicle::createCASTORVehicle(std::string data)
{
	m_Position = 0.0;
	m_RoadPosition = m_Position;

	m_Head		= atoi( data.substr(0,4).c_str() );
	m_Day		= atoi( data.substr(4,2).c_str() );
	m_Month		= atoi( data.substr(6,2).c_str() );
	m_Year		= atoi( data.substr(8,2).c_str() );
	m_Hour		= atoi( data.substr(10,2).c_str() );
	m_Min		= atoi( data.substr(12,2).c_str() );
	m_Sec		= atoi( data.substr(14,2).c_str() );
	m_Hndt		= atoi( data.substr(16,2).c_str() );
	m_Velocity	= atoi( data.substr(18,3).c_str() );
	m_GVW		= atoi( data.substr(21,4).c_str() );
	m_Length	= atoi( data.substr(25,3).c_str() );
	m_NoAxles	= atoi( data.substr(28,1).c_str() );
	m_intDir	= atoi( data.substr(29,1).c_str() );
	m_Lane		= atoi( data.substr(30,1).c_str() );
	m_Trns		= atoi( data.substr(31,3).c_str() );

	m_DirPos = m_intDir == 1 ? true : false;		// Required for compatibility with legacy

	// Length = length/10 for dm to meters 
	// Vel = vel / 10 for dm/s to meters/second
	// GVW * 0.981 for kg/100 to kN
	// Trns = trns/10 for dm to meters

	m_Length	/= 10;
	m_Velocity	/= 10;
	m_GVW		*= KG100_TO_KN;
	m_Trns		/= 10;

	m_Tdelay	= 0.0;
	
	setAxles();
	double W;
	double S;
	int j = 32;

	for(int i = 0; i < m_NoAxles; i++)
	{
		j += 2;
		W = atoi( data.substr(j,3).c_str() );
		W *= KG100_TO_KN;	// convert to kN
		setAW(i, W);
		
		j += 3;
		S = atoi( data.substr(j,2).c_str() );
		S /= 10;		// convert to m
		setAS(i, S);
	}

	m_ID = setID();
}

void Vehicle::createSAFTVehicle(std::string data)
{
	m_Position = 0.0;
	m_RoadPosition = m_Position;

	m_Order		= atoi( data.substr(0,5).c_str() );
	m_Head		= 20005;
	m_Day		= atoi( data.substr(10,2).c_str() );
	m_Month		= atoi( data.substr(12,2).c_str() );
	m_Year		= atoi( data.substr(14,2).c_str() );
	m_Hour		= atoi( data.substr(16,2).c_str() );
	m_Min		= atoi( data.substr(18,2).c_str() );
	m_Sec		= atoi( data.substr(20,2).c_str() );
	m_Hndt		= atoi( data.substr(22,2).c_str() );
	m_Velocity	= atoi( data.substr(24,3).c_str() );
	m_GVW		= atoi( data.substr(27,4).c_str() );
	m_Length	= atoi( data.substr(31,3).c_str() );
	m_NoAxles	= atoi( data.substr(34,1).c_str() );

	m_Lane = 1;
	m_intDir = 1;	// m_DirPointsToRight = true;

	// Length = length/10 for dm to meters 
	// Vel = vel / 10 for dm/s to meters/second
	m_Length	/= 10;
	m_Velocity	/= 10;
	
	m_Tdelay	= 0.0;

	setAxles();
	double W;
	double S;
	int j = 33;

	for(int i = 0; i < m_NoAxles; i++)
	{
		j += 2;
		W = atoi( data.substr(j,3).c_str() );
		setAW(i, W);

		if(i == m_NoAxles-1) break;

		j += 3;
		S = atoi( data.substr(j,2).c_str() );
		S /= 10;		// convert to m
		setAS(i, S);
	}

	m_ID = setID();		// why is this here
}

	/** Prepares a vehicle for printing to a CASTOR file
	 *
	 * @param pVehChar The string of data to represent the truck
	 *
	 * This function takes in a string, and writes all of its properties to
	 * the string in CASTOR format, in order to prepare for printing of the 
	 * vehicle
	 */
void Vehicle::writeCASTORData(char* pVehChar)
{
	// Length = length*10 for meters to decimeters
	// Vel = vel * 10 for metres/second to decimetres/second
	// GVW / 0.981 for KN to KG/100 
	// Trns = trns/10 for metres to decimetres

	int velocity	= Round(m_Velocity*10);
	int grossWeight = Round(m_GVW/KG100_TO_KN);
	int length		= Round(m_Length*10);
	int transPos	= Round(m_Trns*10);

	ostrstream oFile( pVehChar, sizeof(pVehChar)*78 );

	oFile.width(4);	oFile << m_Head;
	oFile.width(2);	oFile << m_Day;
	oFile.width(2);	oFile << m_Month;
	oFile.width(2);	oFile << m_Year;
	oFile.width(2);	oFile << m_Hour;
	oFile.width(2);	oFile << m_Min;
	oFile.width(2);	oFile << m_Sec;
	oFile.width(2);	oFile << m_Hndt;
	oFile.width(3);	oFile << velocity;
	oFile.width(4);	oFile << grossWeight;
	oFile.width(3);	oFile << length;
	oFile.width(1);	oFile << m_NoAxles;
	oFile.width(1);	oFile << m_intDir;
	oFile.width(1);	oFile << getLaneNoInDirection();	// local lane no in direction
	oFile.width(3);	oFile << transPos;
	
	int j = 3;
	for(int i = 0; i < m_NoAxles; i++)
	{
		oFile.width(j);	oFile << Round(this->getAW(i)/KG100_TO_KN);	// convert from kN to kg/100
		j = 2;

		if(i == 8) break;

		oFile.width(j);	oFile << Round(this->getAS(i)*10);		// convert from m to dm
		j = 3;
	}

	for(i = m_NoAxles; i < CASTOR_MAX_AXLES; i++)
	{
		oFile.width(j);	oFile << "0";
		j = 2;

		if(i == 8) break;

		oFile.width(j);	oFile << "0";
		j = 3;
	}

	oFile << ends;
}
	/** Prepares a vehicle for printing to a SAFT file
	 *
	 * @param pVehChar The string of data to represent the truck
	 *
	 * This function takes in a string, and writes all of its properties to
	 * the string in SAFT format, in order to prepare for printing of the 
	 * vehicle
	 */
void Vehicle::writeSAFTData(char *pVehChar)
{
	// Length = length*10 for meters to decimeters
	// Vel = vel * 10 for metres/second to decimetres/second

	int velocity	= Round(m_Velocity*10);
	int length		= Round(m_Length*10);

	ostrstream oFile( pVehChar, sizeof(pVehChar)*78 );

	oFile.fill('0');
	oFile.width(5);	oFile << m_Order;
	oFile.width(5);	oFile << m_Head;
	oFile.width(2);	oFile << m_Day;
	oFile.width(2);	oFile << m_Month;
	oFile.width(2);	oFile << m_Year;
	oFile.width(2);	oFile << m_Hour;
	oFile.width(2);	oFile << m_Min;
	oFile.width(2);	oFile << m_Sec;
	oFile.width(2);	oFile << m_Hndt;
	oFile.fill(' ');
	oFile.width(3);	oFile << velocity;
	oFile.width(4);	oFile << m_GVW;
	oFile.width(3);	oFile << length;
	oFile.width(1);	oFile << m_NoAxles;

	int j = 3;
	for(int i = 0; i < m_NoAxles; i++)
	{
		oFile.width(j);	oFile << Round(this->getAW(i)/KG100_TO_KN);	// convert from kN to kg/100
		j = 2;

		if(i == m_NoAxles-1) 
			break;

		oFile.width(j);	oFile << Round(this->getAS(i)*10);		// convert from m to dm
		j = 3;
	}

	oFile << ends;
}

CString Vehicle::getDataString()
{
	CString dataStr, tempStr, strSplitter;
	strSplitter = "-----------------------------------\n";
	
	dataStr = "Vehicle Data:\n";
	dataStr += strSplitter;
	dataStr += "Direction: \t";
	tempStr = m_DirPos ? "Positive\n" : "Negative\n";
	dataStr += tempStr;
	
	tempStr.Format("Lane:        \t%d\n",			m_Lane);											dataStr += tempStr;
	tempStr.Format("Position:    \t%4.1f m\n",		m_RoadPosition);									dataStr += tempStr;
	tempStr.Format("Velocity:    \t%3.1f km/h\n",	m_Velocity*M_PER_S_TO_KM_PER_H);					dataStr += tempStr;
	tempStr.Format("Acceleration:\t%2.2f m/s^2\n",	m_Acceleration);									dataStr += tempStr;
	tempStr.Format("Desired Vel.:\t%3.1f km/h\n",	m_pDriver->getDesiredVel()*M_PER_S_TO_KM_PER_H);	dataStr += tempStr;
	
	dataStr += strSplitter;
	
	tempStr.Format("IDM T: \t\t%1.2f s\n",		m_IDMDriver.get_T() );								dataStr += tempStr;
	tempStr.Format("IDM A: \t\t%2.2f m/s^2\n",	m_IDMDriver.get_A() );								dataStr += tempStr;
	tempStr.Format("IDM B: \t\t%2.2f m/s^2\n",	m_IDMDriver.get_B() );								dataStr += tempStr;
	tempStr.Format("IDM S0:\t\t%2.2f m\n",		m_IDMDriver.get_S0() );								dataStr += tempStr;
	tempStr.Format("IDM S1:\t\t%2.2f m\n",		m_IDMDriver.get_S1() );								dataStr += tempStr;
	tempStr.Format("IDM V0:\t\t%3.1f m/s\n",	m_IDMDriver.get_V0() );								dataStr += tempStr;
	tempStr.Format("IDM Delta:\t%2.1f\n",		m_IDMDriver.get_Delta() );							dataStr += tempStr;
	tempStr.Format("IDM Polite:\t%2.1f\n",		m_IDMDriver.get_Polite() );							dataStr += tempStr;
	tempStr.Format("IDM Bias:\t%2.1f\n",		m_IDMDriver.get_Bias() );							dataStr += tempStr;
	tempStr.Format("IDM DeltaAth:\t%2.1f\n",	m_IDMDriver.get_DeltaAth() );						dataStr += tempStr;
	
	dataStr += strSplitter;
	
	tempStr.Format("GVW:         \t%3.1f kN\n",		m_GVW);												dataStr += tempStr;
	tempStr.Format("Length:      \t%2.1f m\n",		m_Length);											dataStr += tempStr;
	tempStr.Format("No of Axles: \t%d\n",			m_NoAxles);											dataStr += tempStr;

	for(int i = 0; i < m_NoAxles; i++)
	{
		tempStr.Format("Axle %d:\t", i+1);			dataStr += tempStr;
		tempStr.Format("%2.1f kN",  getAW(i) );	dataStr += tempStr;
		if(i < m_NoAxles - 1)
		{
			tempStr.Format(" | %2.1f m\n",getAS(i) );	dataStr += tempStr;
		}		
	}

	return dataStr;
}
